"""Default value generation for model fields."""

import datetime
import itertools
import string

from django.core.exceptions import ImproperlyConfigured
from django.utils import timezone

try:
    from django.contrib.gis.db import models
except ImproperlyConfigured:
    from django.db import models

try:
    from django.contrib.postgres import fields as postgres
except ImportError:  # psycopg2 is probably not available
    postgres = None


_generators = {}


def register(field_class, generator=None):
    """Register a default value generator for a field class."""

    def decorator(generator):
        _generators[field_class] = generator

    if generator is None:
        return decorator
    decorator(generator)


def generate_default_value(field, unique=False):
    """Generate a default value for a model field."""

    def _default_generator(field, unique=False):
        if unique:
            raise ValueError("unique not supported")
        return None

    generator = _generators.get(type(field), _default_generator)
    return generator(field, unique=unique)


@register(models.BooleanField)
def boolean_field_default(field, unique=False):
    if unique or field.unique:
        raise ValueError("unique not supported")
    return True


def char_field_default(field, unique=False):
    length = min(4, field.max_length or 4)
    if unique or field.unique:
        it = itertools.permutations(string.ascii_letters, length)
        return lambda: "".join(next(it))
    return "" if field.blank else "abcd"[:length]


register(models.CharField, char_field_default)
register(models.FileField, char_field_default)
register(models.ImageField, char_field_default)
register(models.SlugField, char_field_default)
register(models.TextField, char_field_default)


@register(models.DateField)
def date_field_default(field, unique=False):
    if field.auto_now or field.auto_now_add:
        return None
    if unique or field.unique:
        it = itertools.count(step=24 * 60 ** 2)
        return lambda: datetime.date.fromtimestamp(next(it))
    return datetime.date.fromtimestamp(0)


@register(models.DateTimeField)
def date_time_field_default(field, unique=False):
    def todatetime(timestamp):
        t = datetime.datetime.utcfromtimestamp(timestamp)
        return timezone.make_aware(t, timezone.utc)

    if field.auto_now or field.auto_now_add:
        return None
    if unique or field.unique:
        it = itertools.count(step=60 ** 2)
        return lambda: todatetime(next(it))
    return todatetime(0)


@register(models.FloatField)
def float_field_default(field, unique=False):
    if unique or field.unique:
        it = itertools.count()
        return lambda: float(next(it))
    return 0.0


def integer_field_default(field, unique=False):
    if unique or field.unique:
        it = itertools.count()
        return lambda: next(it)
    return 0


register(models.BigIntegerField, integer_field_default)
register(models.DecimalField, integer_field_default)
register(models.IntegerField, integer_field_default)
register(models.PositiveIntegerField, integer_field_default)
register(models.PositiveSmallIntegerField, integer_field_default)
register(models.SmallIntegerField, integer_field_default)


@register(models.TimeField)
def time_field_default(field, unique=False):
    def totime(timestamp):
        return datetime.datetime.utcfromtimestamp(timestamp).time()

    if field.auto_now or field.auto_now_add:
        return None
    if unique or field.unique:
        it = itertools.count()
        return lambda: totime(next(it))
    return totime(0)


def make_geometry_field_default(template):
    def _default(field, unique=False):
        if unique or field.unique:
            it = itertools.count(1)
            return lambda: template % next(it)
        return template % 1

    return _default


for fieldname, template in [
    ("GeometryField", "POINT (%d 0)"),
    ("PointField", "POINT (%d 0)"),
    ("LineStringField", "LINESTRING (%d 0, 0 1)"),
    ("PolygonField", "POLYGON ((0 0, %d 1, 0 1, 0 0))"),
    ("MultiPointField", "MULTIPOINT ((%d 0))"),
    ("MultiLineStringField", "MULTILINESTRING ((%d 0, 0 1))"),
    ("MultiPolygonField", "MULTIPOLYGON (((0 0, %d 1, 0 1, 0 0)))"),
    ("GeometryCollectionField", "GEOMETRYCOLLECTION (POINT (%d 0))"),
]:
    field = getattr(models, fieldname, None)
    if field:
        register(field, make_geometry_field_default(template))


if postgres:

    @register(postgres.JSONField)
    def json_field_default(field, unique=False):
        if unique or field.unique:
            raise ValueError("unique not supported")
        return {}



"""Django database models for pytest-django-ifactory's unit tests."""

from django.contrib.gis.db import models


class ModelA(models.Model):
    name = models.CharField(max_length=64, unique=True)
    category = models.CharField(max_length=64)
    blank = models.CharField(max_length=64, blank=True)


class ModelB(models.Model):
    name = models.CharField(max_length=64, unique=True)
    required_a = models.ForeignKey(ModelA, on_delete=models.CASCADE)
    nullable_a1 = models.ForeignKey(
        ModelA, on_delete=models.CASCADE, null=True, related_name="+"
    )
    nullable_a2 = models.ForeignKey(
        ModelA, on_delete=models.CASCADE, null=True, related_name="+"
    )


class AllFieldsModel(models.Model):
    boolean = models.BooleanField()
    char = models.CharField(max_length=32)
    file = models.FileField()
    image = models.ImageField()
    slug = models.SlugField()
    text = models.TextField()
    date = models.DateField()
    datetime = models.DateTimeField()
    float = models.FloatField()
    biginteger = models.BigIntegerField()
    decimal = models.DecimalField(decimal_places=0, max_digits=9)
    integer = models.IntegerField()
    positiveinteger = models.PositiveIntegerField()
    positivesmallinteger = models.PositiveSmallIntegerField()
    smallinteger = models.SmallIntegerField()
    time = models.TimeField()
    geometry = models.GeometryField()
    point = models.PointField()
    linestring = models.LineStringField()
    polygon = models.PolygonField()
    multipoint = models.MultiPointField()
    multilinestring = models.MultiLineStringField()
    multipolygon = models.MultiPolygonField()
    geometrycollection = models.GeometryCollectionField()



"""Unit and regression tests for the instance factory class."""

import copy

import pytest
from testapp.models import ModelA, ModelB


@pytest.fixture(params=["ifactory", "transactional_ifactory"])
def ifactory(request):
    return request.getfixturevalue(request.param)


def test_all_fields(ifactory):
    instance = ifactory.testapp.allfieldsmodel()
    for field in instance._meta.get_fields():
        assert getattr(instance, field.name) is not None


def test_initialization(ifactory):
    assert hasattr(ifactory, "testapp")
    assert hasattr(ifactory.testapp, "modela")
    assert hasattr(ifactory.testapp, "modelb")
    assert isinstance(ifactory.testapp.modela().name, str)
    b = ifactory.testapp.modelb()
    assert isinstance(b.name, str)
    assert isinstance(b.required_a, ModelA)
    assert b.nullable_a1 is None
    assert b.nullable_a2 is None


def test_blank_default(ifactory):
    a = ifactory.testapp.modela()
    assert not a.blank


def test_create_with_attrs(ifactory):
    assert ifactory.testapp.modela(name="foo").name == "foo"


def test_dynamic_method_names(ifactory):
    assert ifactory.testapp.modela.__name__ == "modela"
    assert ifactory.testapp.modelb.__name__ == "modelb"


class TestConfigure:
    @pytest.fixture
    def cifactory(self, ifactory):
        original_defaults = copy.deepcopy(ifactory._defaults)
        yield ifactory
        type(ifactory)._defaults = original_defaults

    @pytest.mark.parametrize("model", ["testapp.modelb", ModelB])
    def test_primitive(self, cifactory, model):
        cifactory.configure_defaults(model, {"name": "foo"})
        b = cifactory.testapp.modelb()
        assert b.name == "foo"

    def test_callable(self, cifactory):
        cifactory.configure_defaults("testapp.modela", {"name": lambda: "foo"})
        assert cifactory.testapp.modela().name == "foo"

    @pytest.mark.parametrize("model", ["testapp.NoModel", "noapp.NoModel"])
    def test_invalid_model(self, cifactory, model):
        with pytest.raises(LookupError):
            cifactory.configure_defaults(model, {})

    def test_create(self, cifactory):
        cifactory.configure_defaults(
            "testapp.modelb", {"nullable_a1": cifactory.Create}
        )
        assert isinstance(cifactory.testapp.modelb().nullable_a1, ModelA)

    def test_create_with_attrs(self, cifactory):
        cifactory.configure_defaults(
            "testapp.modelb", {"nullable_a1": cifactory.Create(name="foo")}
        )
        assert cifactory.testapp.modelb().nullable_a1.name == "foo"

    def test_lookup(self, cifactory):
        ModelA.objects.create(name="foo")
        ModelA.objects.create(name="bar")
        cifactory.configure_defaults(
            "testapp.modelb", {"nullable_a1": cifactory.Lookup}
        )
        assert cifactory.testapp.modelb().nullable_a1.name == "foo"

    def test_lookup_with_attrs(self, cifactory):
        ModelA.objects.create(name="foo")
        cifactory.configure_defaults(
            "testapp.modelb", {"nullable_a1": cifactory.Lookup(name="foo")}
        )
        assert cifactory.testapp.modelb().nullable_a1.name == "foo"

    def test_unique(self, cifactory):
        cifactory.configure_defaults("testapp.modela", {"category": cifactory.Unique})
        instance1 = cifactory.testapp.modela()
        instance2 = cifactory.testapp.modela()
        assert instance1.category != instance2.category

    def test_plugin(self, cifactory):
        cifactory.configure_defaults("testapp.ModelA", {"name": "adam"})
        cifactory.configure_defaults(
            "testapp.ModelB", {"name": "bert", "nullable_a1": cifactory.Create}
        )

        a = cifactory.testapp.modela(name="alan")
        assert ModelA.objects.count() == 1
        assert a.name == "alan"
        b = cifactory.testapp.modelb(required_a=a)
        assert ModelB.objects.count() == 1
        assert ModelA.objects.count() == 2
        assert b.name == "bert"
        assert b.required_a is a
        assert isinstance(b.nullable_a1, ModelA)
        assert b.nullable_a1.name == "adam"
        assert b.nullable_a2 is None



"""Test configuration for pytest-django-ifactory."""

import os

from django.conf import settings
import py
import pytest

pytest_plugins = ["pytester"]


def pytest_configure():
    settings.configure(
        DATABASES={"default": {"ENGINE": "django.contrib.gis.db.backends.spatialite"}},
        SPATIALITE_LIBRARY_PATH=os.environ.get("SPATIALITE_LIBRARY_PATH"),
        INSTALLED_APPS=[
            "django.contrib.auth",
            "django.contrib.contenttypes",
            "django.contrib.gis",
            "testapp",
        ],
        USE_TZ=True,
    )


@pytest.fixture
def django_testdir(testdir):
    cwd = py.path.local(__file__).dirpath()
    cwd.join("djangosettings.py").copy(testdir.tmpdir)
    cwd.join("testapp").copy(testdir.tmpdir.join("testapp"))
    return testdir



"""A model instance factory for pytest-django."""

import pkg_resources

try:
    distribution = pkg_resources.get_distribution("pytest-django-ifactory")
except pkg_resources.DistributionNotFound:
    pass
else:
    __version__ = distribution.version



"""The Django model instance factory."""

from collections import abc

from django.apps import apps as django_apps
from django.db import models


def generate_default_value(*args, **kwargs):
    # The defaults module's imports use the Django settings so we
    # delay this import as long as possible
    from .defaults import generate_default_value as _generate_default_value

    return _generate_default_value(*args, **kwargs)


class CreateRelatedInstance(object):

    """Flag for creating a related instance using the factory.

    Any *attrs* given to the constructor will be used when creating
    the related instance.

    """

    def __init__(self, **attrs):
        self.attrs = attrs


class LookupRelatedInstance(object):

    """Flag for looking up a related instance in the database.

    The *attrs* given to the constructor must uniquely identify the
    related instance in the database.

    """

    def __init__(self, **attrs):
        self.attrs = attrs


class Unique(object):

    """Flag for generating a unique value for a field."""


class ApplicationNamespaceDescriptor(object):

    """Descriptor for an application namespace on an instance factory.

    Each instance of this class dynamically generates a namespace
    class for the *models* its given.

    """

    def __init__(self, models):
        class ApplicationNamespace(object):
            def __init__(self, ifactory):
                self.ifactory = ifactory

        # We can probably use functools.partialmethod when dropping
        # support for Python 2.7.
        def make_create(model):
            def create(self, **attrs):
                return self.ifactory.create(model, attrs)

            create.__name__ = model._meta.model_name
            return create

        for model in models:
            setattr(ApplicationNamespace, model._meta.model_name, make_create(model))

        self.namespace_class = ApplicationNamespace

    def __get__(self, instance, owner):
        return self.namespace_class(instance)


class InstanceFactory(object):

    """A factory for Django model instances to use in unit tests."""

    Create = CreateRelatedInstance
    Lookup = LookupRelatedInstance
    Unique = Unique

    # Global defaults for all models in the format
    # {model_label_lower: {field_name: default_value, ...}, ...}.
    _defaults = {}

    @classmethod
    def register_all_models(cls):
        """Register all installed models for production at the factory."""
        if not django_apps.ready:
            raise RuntimeError(
                "InstanceFactory.register_models() must be called after Django has "
                "been initialized"
            )
        for app in django_apps.get_app_configs():
            models = list(app.get_models())
            for model in models:
                cls._set_initial_defaults(model)
            setattr(cls, app.label, ApplicationNamespaceDescriptor(models))

    @classmethod
    def _set_initial_defaults(cls, model):
        # Generate initial default values for the fields of a *model*.
        def needs_default(field):
            return (
                not field.null
                and not field.has_default()
                and not isinstance(field, models.AutoField)
                and not field.many_to_many
                and not field.one_to_many
            )

        def generate_default(field):
            if field.many_to_one or field.one_to_one:
                return InstanceFactory.Create
            else:
                return generate_default_value(field)

        cls._defaults[model._meta.label_lower] = {
            field.name: generate_default(field)
            for field in model._meta.get_fields()
            if needs_default(field)
        }

    @classmethod
    def configure_defaults(cls, model, attrs):
        """Configure the default values for the given *model*.

        *model* can be an actual model class or a model label. *attrs*
        should map field names to default values.

        """

        def isunique(value):
            return value is cls.Unique or isinstance(value, cls.Unique)

        if isinstance(model, str):
            model = django_apps.get_model(model)
        if not cls._defaults:  # make sure the factory is initialized
            cls.register_all_models()
        cls._defaults[model._meta.label_lower].update(
            {
                name: generate_default_value(model._meta.get_field(name), unique=True)
                if isunique(value)
                else value
                for name, value in attrs.items()
            }
        )

    def __init__(self):
        if not self._defaults:  # make sure the factory is initialized
            self.register_all_models()

    def create(self, model, attrs=()):
        """Create an instance of the given model.

        This is the generic and more low-level method to create model
        instances. It can be useful if the model is not known before
        runtime. Otherwise, use the ifactory.applabel.modelname()
        shortcuts for this method.

        """

        def get_default_value(field, value):
            if value is self.Create:
                value = self.create(field.related_model)
            elif isinstance(value, self.Create):
                value = self.create(field.related_model, value.attrs)
            elif value is self.Lookup:
                value = field.related_model.objects.first()
            elif isinstance(value, self.Lookup):
                value = field.related_model.objects.get(**value.attrs)
            elif isinstance(value, abc.Callable):
                value = value()
            return value

        defaults = dict(
            (k, get_default_value(model._meta.get_field(k), v))
            for k, v in self._defaults[model._meta.label_lower].items()
            if k not in attrs
        )
        defaults.update(attrs)
        instance = model(**defaults)
        instance.clean()
        instance.save()
        return instance



# Generated by Django 2.1 on 2018-08-26 01:14

import django.contrib.gis.db.models.fields
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='AllFieldsModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('boolean', models.BooleanField()),
                ('char', models.CharField(max_length=32)),
                ('file', models.FileField(upload_to='')),
                ('image', models.ImageField(upload_to='')),
                ('slug', models.SlugField()),
                ('text', models.TextField()),
                ('date', models.DateField()),
                ('datetime', models.DateTimeField()),
                ('float', models.FloatField()),
                ('biginteger', models.BigIntegerField()),
                ('decimal', models.DecimalField(decimal_places=0, max_digits=9)),
                ('integer', models.IntegerField()),
                ('positiveinteger', models.PositiveIntegerField()),
                ('positivesmallinteger', models.PositiveSmallIntegerField()),
                ('smallinteger', models.SmallIntegerField()),
                ('time', models.TimeField()),
                ('geometry', django.contrib.gis.db.models.fields.GeometryField(srid=4326)),
                ('point', django.contrib.gis.db.models.fields.PointField(srid=4326)),
                ('linestring', django.contrib.gis.db.models.fields.LineStringField(srid=4326)),
                ('polygon', django.contrib.gis.db.models.fields.PolygonField(srid=4326)),
                ('multipoint', django.contrib.gis.db.models.fields.MultiPointField(srid=4326)),
                ('multilinestring', django.contrib.gis.db.models.fields.MultiLineStringField(srid=4326)),
                ('multipolygon', django.contrib.gis.db.models.fields.MultiPolygonField(srid=4326)),
                ('geometrycollection', django.contrib.gis.db.models.fields.GeometryCollectionField(srid=4326)),
            ],
        ),
        migrations.CreateModel(
            name='ModelA',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=64, unique=True)),
                ('category', models.CharField(max_length=64)),
                ('blank', models.CharField(blank=True, max_length=64)),
            ],
        ),
        migrations.CreateModel(
            name='ModelB',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=64, unique=True)),
                ('nullable_a1', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='testapp.ModelA')),
                ('nullable_a2', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='testapp.ModelA')),
                ('required_a', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='testapp.ModelA')),
            ],
        ),
    ]



"""Build and install pytest-django-ifactory."""

from setuptools import setup, find_packages

with open("README.md", "r") as stream:
    long_description = stream.read()

setup(
    name="pytest-django-ifactory",
    version="0.4.0",
    author="Mattias Jakobsson",
    author_email="mjakob422@gmail.com",
    description="A model instance factory for pytest-django",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/gorilladev/pytest-django-ifactory",
    license="BSD-3-Clause",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Plugins",
        "Framework :: Django",
        "Framework :: Pytest",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Testing",
    ],
    keywords="pytest django database testing",
    packages=find_packages(exclude=["tests*"]),
    install_requires=["Django", "pytest-django"],
    entry_points={"pytest11": ["django-ifactory = pytest_django_ifactory.plugin"]},
)



"""Pytest plugin module."""

import pytest

from .ifactory import InstanceFactory


class PytestDjangoIFactorySpec:

    """Hook specification namespace for this plugin."""

    def pytest_django_ifactory_configure(self, ifactory):
        """Configure model field defaults at the pytest instance factory."""


def pytest_addhooks(pluginmanager):
    pluginmanager.add_hookspecs(PytestDjangoIFactorySpec)


@pytest.fixture(autouse=True, scope="session")
def _django_ifactory_register(request):
    request.config.pluginmanager.hook.pytest_django_ifactory_configure(
        ifactory=InstanceFactory
    )


@pytest.fixture
def ifactory(db):
    return InstanceFactory()


@pytest.fixture
def transactional_ifactory(transactional_db):
    return InstanceFactory()
